/*
 * Copyright 2009 Dan Kegel
 * Copyright 2009 Jacek Caban for Codeweavers
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

#include "wine/test.h"

static char curdir[2048];

/*
 * Return TRUE if cmd.exe is available
 */
static int cmd_available(void)
{
    STARTUPINFOA si;
    PROCESS_INFORMATION pi;

    memset(&si, 0, sizeof(si));
    si.cb = sizeof(si);
    if (CreateProcessA(NULL, "cmd /c exit 0", NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) {
        CloseHandle(pi.hThread);
        CloseHandle(pi.hProcess);
        return TRUE;
    }
    return FALSE;
}

/* Run test_name.cmd, save stdout to test_name.out, return TRUE if could create process.
 * Throws away stderr for now, since we don't want to match Windows error messages
 * word-for-word unless some app depends on them.
 * Returns exit code in *pExitCode.
 */
static BOOL run_one_test(const char *test_name, DWORD *pExitCode)
{
    SECURITY_ATTRIBUTES sa;
    STARTUPINFOA si;
    PROCESS_INFORMATION pi;
    HANDLE file;
    const char *res;
    DWORD size;
    BOOL bres = FALSE;
    char cmdname[2048];
    char outname[2048];

    trace("run_test(%s)\n", test_name);
    sprintf(cmdname, "%s.cmd", test_name);
    sprintf(outname, "%s.out", test_name);

    memset(&sa, 0, sizeof(sa));
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = NULL;
    sa.bInheritHandle = TRUE;

    file = CreateFileA(outname, GENERIC_WRITE, 0, &sa, CREATE_ALWAYS,
            FILE_ATTRIBUTE_NORMAL, NULL);
    if (file == INVALID_HANDLE_VALUE) {
        skip("can't create output file %s, skipping cmd test\n", outname);
        return FALSE;
    }

    memset(&si, 0, sizeof(si));
    si.cb = sizeof(si);
    si.dwFlags = STARTF_USESTDHANDLES;
    /* Should we dup these? */
    si.hStdInput = GetStdHandle(STD_INPUT_HANDLE);
    si.hStdOutput = file;
    si.hStdError = GetStdHandle(STD_ERROR_HANDLE);

    bres = CreateProcessA(NULL, cmdname, NULL, NULL, TRUE, 0, NULL, NULL, &si, &pi);
    CloseHandle(file);
    if (!bres) {
        skip("can't create process '%s', skipping cmd test\n", cmdname);
        return FALSE;
    }

    WaitForSingleObject(pi.hProcess, INFINITE);
    if (!GetExitCodeProcess(pi.hProcess, pExitCode))
        trace("GetExitCodeProcess failed?\n");
    else
        trace("Exit code for %s is %d\n", cmdname, *pExitCode);
    CloseHandle(pi.hThread);
    CloseHandle(pi.hProcess);
    return TRUE;
}

/* Remove trailing whitespace.  Only needed until next release of Wine. */
static void chomp(char *buf)
{
    char *p = buf + strlen(buf) - 1;
    while ((p >= buf) && isspace(*p))
       *p-- = 0;
}

/* Compare given output to reference, return 0 on identical, nonzero on different */
/* FIXME: don't quit after first error; skip to next test in file */
static int check(const char *name, const char *suffix)
{
    char fname_out[2048];
    char fname_exp[2048];
    FILE *fp_out = NULL, *fp_exp = NULL;
    char buf_out[2048];
    char buf_exp[2048];
    int line;
    int ret = 1;

    sprintf(fname_out, "%s.%s", name, suffix);
    fp_out = fopen(fname_out, "rb");

    sprintf(fname_exp, "%s.%s.exp", name, suffix);
    fp_exp = fopen(fname_exp, "rb");

    if (fp_exp && ! fp_out) {
        trace("%s exists, but %s doesn't\n", fname_exp, fname_out);
        goto out;
    }
    if (!fp_exp && fp_out) {
        trace("%s exists, but %s doesn't\n", fname_out, fname_exp);
        goto out;
    }

    line = 0;
    while (1) {
        const char *pout, *pexp;
        char *p;

        line++;
        pout = fgets(buf_out, sizeof(buf_out), fp_out);
        pexp = fgets(buf_exp, sizeof(buf_exp), fp_exp);

        if (!pout && !pexp)
            break;

        if (pexp && ! pout) {
            trace("%s EOF before %s\n", fname_out, fname_exp);
            goto out;
        }
        if (!pexp && pout) {
            trace("%s EOF before %s\n", fname_exp, fname_out);
            goto out;
        }
        p = buf_exp;
        while (p = strstr(p, "@@")) {
            /* expand @@ to current directory */
            memmove(p + strlen(curdir), p + 2, strlen(p+2)+1);
            memcpy(p, curdir, strlen(curdir));
        }
        /* trim trailing CR or LF's to work around another bug */
        chomp(buf_exp);
        chomp(buf_out);

        if (strcmp(pout, pexp)) {
            trace("%s and %s differ on line %d: '%s' vs '%s'\n", fname_out, fname_exp, line, pout, pexp);
            goto out;
        }
    }
    ret= 0;
out:
    if (fp_out)
        fclose(fp_out);
    if (fp_exp)
        fclose(fp_exp);
    return ret;
}

static void run_tests(void)
{
    WIN32_FIND_DATA finddata;
    HANDLE hFind;

    /* Run all test*.cmd files in current directory, compare output to test*.out.exp */

    hFind = FindFirstFile("test*.cmd", &finddata);
    if (INVALID_HANDLE_VALUE == hFind) {
        printf("no tests?\n");
        return;
    }
    do {
       char testname[2048];
       const char *dotcmd = strrchr(finddata.cFileName, '.');
       int len;
       DWORD err;
       int pass;

       if (!dotcmd) {
           printf("bug\n");
           continue;
       }
       len = dotcmd - finddata.cFileName;
       memcpy(testname, finddata.cFileName, len);
       testname[len] = 0;

       err = 0;
       pass = (run_one_test(testname, &err) && !check(testname, "out") && err == 0);
       /* TODO: apply todo_wine for tests in files named *-todo.cmd */
       /* TODO: provide more informative error message without relying on trace() in run_one_test() */
       ok(pass, "test %s\n", testname);
   }
   while (FindNextFile(hFind, &finddata) != 0);
}

static BOOL CALLBACK unpack_file_proc(HMODULE hModule, LPCTSTR lpszType,
                                      LPTSTR lpszName, LONG_PTR lParam)
{
    const char *dir=(const char *)lParam;
    DWORD size;
    FILE *fp;
    HGLOBAL hdl;
    HRSRC rsrc;
    LPVOID addr;
    char *buf;
    char fname[2048];
    int nwritten;

    if (!(rsrc = FindResourceA(NULL, lpszName, lpszType)) ||
        !(size = SizeofResource(0, rsrc)) ||
        !(hdl = LoadResource(0, rsrc)) ||
        !(addr = LockResource(hdl)))
        return FALSE;

    /* note: dir must end in a path separator.  It does if it comes from GetTempPath(). */
    sprintf(fname, "%s%s", dir, lpszName);
    fp = fopen(fname, "wb");
    if (!fp) {
       perror(fname);
       return FALSE;
    }
    nwritten = fwrite(addr, 1, size, fp);
    if (fclose(fp) || size != nwritten) {
       perror(fname);
       return FALSE;
    }
    return TRUE;
}

/* Unpack all the TESTDATA resources to the given directory */
int extract_test_data(const char *dir)
{
    return EnumResourceNames (NULL, "TESTDATA", unpack_file_proc, (LPARAM)dir);
}

START_TEST(cmd)
{
    int interactive;

    if (!cmd_available()) {
        skip("cmd not installed, skipping cmd tests\n");
        return;
    }

    /* Developers can use interactive mode to avoid having to
     * recompile this program every time they change a test.
     */
    interactive = (getenv("WINETEST_INTERACTIVE") != NULL);


    if (!interactive) {
        char tempdir[2048];

        GetTempPath(sizeof(tempdir), tempdir);

        /* If not interactive, create an empty temporary directory,
         * cd to it, and unpack the bundled tests there.
         * FIXME: just work in system temp directory for now.
         */
        SetCurrentDirectory(tempdir);
        system("erase TEST*.* > nul 2> nul");

        if (!extract_test_data(tempdir)) {
           ok(FALSE, "extraction failed\n");
           return;
        }
    }

    /* Run the tests.  */
    GetCurrentDirectory(sizeof(curdir), curdir);
    if (curdir[strlen(curdir)-1] != '\\')
        strcat(curdir, "\\");
    trace("Running tests in directory %s\n", curdir);
    run_tests();

    if (!interactive) {
        /* Remove our temporary directory. 
         * FIXME: just remove our files from system temp directory for now.
         */
        system("erase TEST*.*");
    }
}
